#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined(_MSC_VER)
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef signed short int16_t;
typedef unsigned int uint32_t;
typedef signed int int32_t;
#else
#include <inttypes.h>
#endif

#define CODE_CLEAR 256
#define CODE_END 257

#define MAX_WIDTH 2048
#define MAX_HEIGHT 2048

typedef struct
{
    uint8_t* data;
    unsigned size;
} data_t;
    
#define DICT_SIZE 4096

typedef struct
{
    data_t entries[DICT_SIZE];
    unsigned size;
} lzw_dict_t;

typedef struct
{
    uint16_t code;
    uint16_t dict_size;
} code_output_t;

typedef struct
{
    uint8_t r;
    uint8_t g;
    uint8_t b;
} color_t;
     
typedef struct
{
    uint32_t num_colors;
    color_t palette[256];
    uint32_t width;
    uint32_t height;
    uint8_t* data;
} image_t;


#pragma pack(1)
typedef struct
{
    uint16_t magic;
    uint32_t file_size;
    uint16_t unused_06;
    uint16_t unused_08;
    uint32_t bitmap_data_offset;
    uint32_t header_size; /* from this point */
    int32_t width;
    int32_t height;
    uint16_t num_planes;
    uint16_t bpp;
    uint32_t compression; /* 0 = no compression */
    uint32_t size_of_raw_bmp_data;
    int32_t horizontal_pixels_per_meter;
    int32_t vertical_pixels_per_meter;
    uint32_t num_colors_in_palette;
    uint32_t num_important_colors;
} bmp_header_t;

typedef struct
{
    uint16_t width;
    uint16_t height;

    uint8_t color_table_size:3;
    uint8_t sort_flag:1;
    uint8_t color_resolution:3;
    uint8_t gct_flag:1;
    
    uint8_t background_color;
    uint8_t default_aspect_ratio; /* set to 0 */
} gif_lsd_t;

typedef struct
{
    uint8_t introducer; /* 0x21 */
    uint8_t label; /* 0xf9 */
    uint8_t size; /* 0x04 */

    uint8_t transparent_color_flag:1;
    uint8_t user_input_flag:1;
    uint8_t disposal_method:1;
    uint8_t reserved:3;

    uint16_t delay_time;
    uint8_t transparent_color_index;
    uint8_t terminator; /* 0x00 */
} gif_gce_t;

typedef struct
{
    uint8_t introducer; /* 0x2c */
    uint16_t left_pos;
    uint16_t top_pos;
    uint16_t width;
    uint16_t height;

    uint8_t size_of_local_color_table:3;
    uint8_t reserved:2;
    uint8_t sort_flag:1;
    uint8_t interlace_flag:1;
    uint8_t local_color_table_flag:1;
} gif_id_t;
#pragma pack()

int init_data( data_t* data, unsigned size )
{
    if( data->data )
	free(data->data);
    data->data =  (uint8_t*)malloc(size);
    data->size = size;
    return 1;
}

int set_data( data_t* data, uint8_t* bytes, unsigned size )
{
    if( !init_data(data, size) )
	return 0;
    memcpy( data->data, bytes, size );
    return 1;
}

void dealloc_data( data_t* data )
{
    if( data->data )
	free(data->data);
    data->data = NULL;
    data->size = 0;
}

int append_data( data_t* data, uint8_t c )
{
    data->data = (uint8_t*)realloc(data->data, data->size + 1 );
    if( !data->data )
	return 0;
    data->data[data->size] = c;
    data->size++;
    return 1;
}

image_t* alloc_image( uint32_t num_colors, uint32_t width, uint32_t height )
{
    image_t* image = (image_t*)calloc( 1, sizeof(image_t) );
    if( !image )
	return NULL;
    image->data = (uint8_t*)calloc( 1, width * height );
    if( !image->data )
    {
	free(image);
	return NULL;
    }

    image->num_colors = num_colors;
    image->width = width;
    image->height = height;
    return image;
}

void free_image( image_t* image )
{
    if( image->data )
	free(image->data);
    free(image);
}

image_t* read_bmp( const char* filename )
{
    FILE* fp = NULL;
    bmp_header_t header;
    image_t* image = NULL;
    uint32_t i;
    uint32_t row;
    uint32_t bytes_per_line;
    unsigned num_colors;
    fp = fopen( filename, "rb");
    if( !fp )
	return NULL;

    /* read header */
    if( fread( &header, sizeof(header), 1, fp ) != 1 )
	goto error;

    /* parse header */
    if( header.width < 0 || header.width > MAX_WIDTH )
	goto error;
    if( header.height < 0 || header.height > MAX_HEIGHT )
	goto error;
    if( header.bpp != 8 )
	goto error;
    if( header.compression != 0 )
	goto error;

    num_colors = header.num_colors_in_palette;
    if( num_colors == 0 )
	num_colors = 256;
       
    image = alloc_image( num_colors, header.width, header.height );
    if( !image )
	goto error;
    
    bytes_per_line = (image->width + 3) & ~3;

    
    /* read palette */
    for( i = 0; i < num_colors; i++ )
    {
	uint8_t buf[4];
	if( fread( buf, sizeof(buf), 1, fp ) != 1 )
	    goto error;
	
	image->palette[i].r = buf[2];
	image->palette[i].g = buf[1];
	image->palette[i].b = buf[0];
    }

    /* read data */
    for( row = 0; row < image->height; row++ )
    {
	if( fseek( fp, header.bitmap_data_offset + (image->height-row-1) * bytes_per_line, SEEK_SET ) != 0 )
	    goto error;
	if( fread( &image->data[row * image->width], image->width, 1, fp ) != 1 )
	    goto error;
    }

    fclose(fp);
    return image;
error:
    if( image )
	free( image );

    if( fp )
	fclose( fp );
    return NULL;
}

int write_gif_magic( FILE* fp )
{
    const char magic[] = "GIF89a";
    if( fwrite(magic, strlen(magic), 1, fp ) != 1 )
	return 0;
    return 1;
}

int write_gif_lsd( FILE* fp, image_t* image )
{
    gif_lsd_t lsd;
    memset( &lsd, 0, sizeof(lsd) );

    lsd.width = image->width;
    lsd.height = image->height;
    lsd.gct_flag = 1;
    lsd.color_resolution = 7;
    lsd.sort_flag = 0;
    lsd.color_table_size = 7; /* 2^(7+1) */
    lsd.background_color = 0;
    lsd.default_aspect_ratio = 0;

    if( fwrite( &lsd, sizeof(lsd), 1, fp ) != 1 )
	return 0;

    return 1;
}

int write_gif_color_table( FILE* fp, image_t* image )
{
    unsigned i;
    for( i = 0; i < 256; i++ )
    {
	if( fwrite(&image->palette[i].r, 1, 1, fp ) != 1 )
	    return 0;
	if( fwrite(&image->palette[i].g, 1, 1, fp ) != 1 )
	    return 0;
	if( fwrite(&image->palette[i].b, 1, 1, fp ) != 1 )
	    return 0;
    }
    return 1;
}


int write_gif_gce( FILE* fp, image_t* image )
{
    gif_gce_t gce;
    memset( &gce, 0, sizeof(gce) );

    gce.introducer = 0x21;
    gce.label = 0xf9;
    gce.size = 0x04;
    gce.reserved = 0;
    gce.disposal_method = 0;
    gce.user_input_flag = 0;
    gce.transparent_color_flag = 0;
    gce.delay_time = 0;
    gce.transparent_color_index = 0xff;
    gce.terminator = 0;

    if( fwrite( &gce, sizeof(gce), 1, fp ) != 1 )
	return 0;

    return 1;
}

int write_gif_id( FILE* fp, image_t* image )
{
    gif_id_t id;
    memset( &id, 0, sizeof(id) );

    id.introducer = 0x2c;
    id.left_pos = 0;
    id.top_pos = 0;
    id.width = image->width;
    id.height = image->height;
    id.local_color_table_flag = 0;
    id.interlace_flag = 0;
    id.sort_flag = 0;
    id.reserved = 0;
    id.size_of_local_color_table = 0;

    if( fwrite( &id, sizeof(id), 1, fp ) != 1 )
	return 0;

    return 1;
}

int write_gif_terminator( FILE* fp )
{
    uint8_t terminator = 0x3b;

    if( fwrite( &terminator, 1, 1, fp ) != 1 )
	return 0;

    return 1;
}

void clear_dict( lzw_dict_t* dict )
{
    unsigned i;
    for( i = 0; i < DICT_SIZE; i++ )
	dealloc_data( &dict->entries[i] );
    dict->size = 0;
}

void free_dict( lzw_dict_t* dict )
{
    clear_dict( dict );
    free( dict );
}

unsigned add_dict_entry( lzw_dict_t* dict, uint8_t* data, unsigned size )
{
    unsigned i = dict->size;
    if( i >= DICT_SIZE )
	return ~0;

    if( data && size > 0 )
    {
	if( !set_data( &dict->entries[i], data, size ) )
	    return ~0;
    }
    else
    {
	dict->entries[i].size = 0;
	dict->entries[i].data = NULL;
    }
    dict->size++;
    return i;
}

void init_dict( lzw_dict_t* dict )
{
    unsigned i;
    
    clear_dict( dict );
    for( i = 0; i < 256; i++ )
    {
	uint8_t v = i;
	add_dict_entry( dict, &v, 1 );
    }
    add_dict_entry( dict, NULL, 0 );
    add_dict_entry( dict, NULL, 0 );
}

lzw_dict_t* alloc_dict()
{
    return (lzw_dict_t*)calloc(1,sizeof(lzw_dict_t));
}

unsigned find_dict_entry( lzw_dict_t* dict, data_t* data )
{
    unsigned i;
    for( i = 0; i < dict->size; i++ )
    {
	if( data->size &&
	    data->size == dict->entries[i].size &&
	    dict->entries[i].size &&
	    memcmp( dict->entries[i].data, data->data, data->size ) == 0 )
	    return i;
    }
    return ~0;
}

int is_in_dict( lzw_dict_t* dict, data_t* data, uint8_t last )
{
    unsigned i;
    for( i = 0; i < dict->size; i++ )
    {
	if( data->size + 1 == dict->entries[i].size &&
	    memcmp( dict->entries[i].data, data->data, data->size ) == 0 &&
	    dict->entries[i].data[data->size] == last )
	    return 1;
    }
    return 0;
}


code_output_t* get_lzw_codes( image_t* image )
{
    lzw_dict_t* dict;
    unsigned i;
    data_t w;
    code_output_t* codes;
    unsigned oi;
    memset( &w, 0, sizeof(w) );
    codes = (code_output_t*)calloc( image->width*image->height * 2, sizeof(code_output_t) ); // twice the size is enough :)
    oi = 0;

    
    dict = alloc_dict();
    if( !dict )
	return NULL;
    init_dict( dict );

    codes[oi].code = CODE_CLEAR;
    codes[oi].dict_size = dict->size;
    oi++;
    
    for( i = 0; i < image->width*image->height; i++ )
    {
	uint8_t c = image->data[i];
					     
	if( is_in_dict( dict, &w, c ) )
	{
	    append_data( &w, c );
	}
	else
	{
	    codes[oi].code = find_dict_entry( dict, &w );
	    codes[oi].dict_size = dict->size;
	    oi++;
	    append_data( &w, c );

	    if( dict->size == DICT_SIZE )
	    {
		codes[oi].code = CODE_CLEAR;
		codes[oi].dict_size = dict->size+1;
		oi++;
		init_dict( dict );
	    }
	    else
	    {
		add_dict_entry( dict, w.data, w.size );
	    }
	    set_data( &w, &c, 1);
	}
    }
    codes[oi].code = find_dict_entry( dict, &w );
    codes[oi].dict_size = dict->size;
    oi++;
    codes[oi].code = CODE_END;
    codes[oi].dict_size = dict->size;
    oi++;

    free_dict(dict);
    return codes;
}

unsigned get_bits_needed( unsigned dict_size )
{
    if( dict_size > 2048 )
	return 12;
    if( dict_size > 1024 )
	return 11;
    if( dict_size > 512 )
	return 10;
    return 9;
}

int write_gif_data( FILE* fp, image_t* image )
{
    code_output_t* codes = NULL;
    code_output_t* c;

    uint8_t* buf = NULL;
    uint8_t* p;
    unsigned bit_counter = 0;
    uint8_t size;
    uint8_t min_code_size;
    unsigned i;
    unsigned bytes_to_write;

    codes = get_lzw_codes( image );
    if( !codes )
	goto error;

    buf = (uint8_t*)calloc( 1, image->width * image->height * 2 );
    if( !buf )
	goto error;
    
    min_code_size = 8;
    if( fwrite( &min_code_size, 1, 1, fp ) != 1 )
	goto error;

    c = codes;
    for(;;)
    {
	unsigned bits_needed = get_bits_needed( c->dict_size );
	unsigned code = c->code;

	/* fill bits, lsb first */
	for( i = 0; i < bits_needed; i++ )
	{
	    if( code & 1 )
		buf[bit_counter>>3] |= 1<<(bit_counter & 7);
	    bit_counter++;
	    code >>= 1;
	}
	if( c->code == CODE_END )
	    break;
	c++;
    } 

    p = buf;
    bytes_to_write = (bit_counter+7) >> 3;
    while( bytes_to_write > 0 )
    {
	if( bytes_to_write > 255 )
	    size = 255;
	else
	    size = bytes_to_write;
	if( fwrite( &size, 1, 1, fp ) != 1 )
	    goto error;
	if( fwrite( p, size, 1, fp ) != 1 )
	    goto error;
	bytes_to_write -= size;
	p += size;
    }

    /* final block */
    size = 0;
    if( fwrite( &size, 1, 1, fp ) != 1 )
	goto error;

    free( codes );
    free( buf );
    return 1;
error:
    if( codes )
	free( codes );
    if( buf )
	free( buf );
    return 0;
}

int write_gif( image_t* image, const char* filename )
{
    FILE* fp = NULL;

    fp = fopen(filename, "wb");
    if( !fp )
	goto error;

    if( !write_gif_magic( fp ) )
	goto error;

    if( !write_gif_lsd( fp, image ) )
	goto error;

    if( !write_gif_color_table( fp, image ) )
	goto error;

    if( !write_gif_gce( fp, image ) )
	goto error;

    if( !write_gif_id( fp, image ) )
	goto error;

    if( !write_gif_data( fp, image ) )
	goto error;

    if( !write_gif_terminator( fp ) )
	goto error;
    
    fclose(fp);
    return 1;
error:
    if( fp )
	fclose(fp);
    return 0;
}

int main(int argc, char** argv )
{
    image_t* image;
    if( argc != 3 )
	return 1;

    image = read_bmp( argv[1] );
    if( !image )
	return 2;

    if( !write_gif( image, argv[2] ) )
    {
	free_image(image);
	return 3;
    }
    
    if( image )
	free_image(image);
    return 0;
}
